mat4 unpackMatrix(float id)
{
	vec4 a1,a2,a3,a4;
	a1 = texture2D(BoneMatrixArray,vec2( (id+0.5)/BoneNumberM4 , 3.5/4.0 )).rgba;
	a2 = texture2D(BoneMatrixArray,vec2( (id+0.5)/BoneNumberM4 , 2.5/4.0 )).rgba;
	a3 = texture2D(BoneMatrixArray,vec2( (id+0.5)/BoneNumberM4 , 1.5/4.0 )).rgba;
	a4 = texture2D(BoneMatrixArray,vec2( (id+0.5)/BoneNumberM4 , 0.5/4.0 )).rgba;
	return mat4(a1,a2,a3,a4);
}

//mat4 unpackMatrix(int id)
//{
	//vec4 a,b,c,d;
	//int pos = id * 3;
	//a = texelFetchBuffer(tBoneTBO, pos+0).rgba;
	//b = vec4(0.0,1.0,0.0,0.0);
	//c = vec4(0.0,0.0,1.0,0.0);
	//d = vec4(0.0,0.0,0.0,1.0);
	//b = texelFetchBuffer(tBoneTBO, pos+1).abgr;
	//c = texelFetchBuffer(tBoneTBO, pos+2).abgr;
	//d = texelFetch(tBoneTBO, pos+3).abgr;
	//return mat4(a,b,c,d);
//}
	
vec4 matrixToQuaternion(mat4 m4)
{
	float m00=m4[0][0],m01=m4[0][1],m02=m4[0][2]
	     ,m10=m4[1][0],m11=m4[1][1],m12=m4[1][2]
		 ,m20=m4[2][0],m21=m4[2][1],m22=m4[2][2];
	float t = m00 + m11 + m22;
	vec4 result=vec4(0,0,0,1);
	float x,y,z,w;
	if (t >= 0) 
	{ 
		float s = sqrt(t + 1); 
		w = 0.5f * s;
		s = 0.5f / s;                 
		x = (m21 - m12) * s;
		y = (m02 - m20) * s;
		z = (m10 - m01) * s;
	} 
	else if ((m00 > m11) && (m00 > m22)) 
	{
		float s = sqrt(1.0f + m00 - m11 - m22); 
		x = s * 0.5f; 
		s = 0.5f / s;
		y = (m10 + m01) * s;
		z = (m02 + m20) * s;
		w = (m21 - m12) * s;
	} 
	else if (m11 > m22) 
	{
		float s = sqrt(1.0f + m11 - m00 - m22); 
		y = s * 0.5f; 
		s = 0.5f / s;
		x = (m10 + m01) * s;
		z = (m21 + m12) * s;
		w = (m02 - m20) * s;
	} 
	else 
	{
		float s = sqrt(1.0f + m22 - m00 - m11); 
		z = s * 0.5f; 
		s = 0.5f / s;
		x = (m02 + m20) * s;
		y = (m21 + m12) * s;
		w = (m10 - m01) * s;
	}
	result.x=x;
	result.y=y;
	result.z=z;
	result.w=w;
	return result;
}

mat4 quaternionToMatrix(vec4 quat)
{
	mat4 result=mat4(1.0);
	float x=quat.x,y=quat.y,z=quat.z,w=quat.w;
	float norm = w * w + x * x + y * y + z * z;
	float s = (norm == 1.0) ? 2.0 : (norm > 0.0) ? 2.0 / norm : 0.0;
	float xs = x * s;
	float ys = y * s;
	float zs = z * s;
	float xx = x * xs;
	float xy = x * ys;
	float xz = x * zs;
	float xw = w * xs;
	float yy = y * ys;
	float yz = y * zs;
	float yw = w * ys;
	float zz = z * zs;
	float zw = w * zs;
	result[0][0] = 1 - (yy + zz);
	result[0][1] = (xy - zw);
	result[0][2] = (xz + yw);
	result[1][0] = (xy + zw);
	result[1][1] = 1 - (xx + zz);
	result[1][2] = (yz - xw);
	result[2][0] = (xz - yw);
	result[2][1] = (yz + xw);
	result[2][2] = 1 - (xx + yy);
	return result;
}

vec4 slerp(vec4 q1,vec4 q2,float t)
{
	if (q1.x == q2.x && q1.y == q2.y && q1.z == q2.z && q1.w == q2.w) 
	{
		return q1;
	}

	float result = (q1.x * q2.x) + (q1.y * q2.y) + (q1.z * q2.z)+ (q1.w * q2.w);
	if (result < 0.0) 
	{
		q2.x = -q2.x;
		q2.y = -q2.y;
		q2.z = -q2.z;
		q2.w = -q2.w;
		result = -result;
	}
	float scale0 = 1 - t;
	float scale1 = t;
	if ((1 - result) > 0.1) 
	{
		float theta = acos(result);
		float invSinTheta = 1.0 / sin(theta);
		scale0 = sin((1 - t) * theta) * invSinTheta;
		scale1 = sin((t * theta)) * invSinTheta;
	}
	vec4 resultQuat=vec4(0,0,0,1);
	resultQuat.x = (scale0 * q1.x) + (scale1 * q2.x);
	resultQuat.y = (scale0 * q1.y) + (scale1 * q2.y);
	resultQuat.z = (scale0 * q1.z) + (scale1 * q2.z);
	resultQuat.w = (scale0 * q1.w) + (scale1 * q2.w);
	return resultQuat;
}
vec4 getPosition()
{
	mat4 tempMatrix;
	mat4 my_BoneMatrix0,my_BoneMatrix1,my_BoneMatrix2,my_BoneMatrix3;
	my_Vertex = vec4(vPosition+vFace,1.0);
	vertex.normal = vNormal;
	vertex.vTexCoord = vUV;
	//boneWeightType=0;
	if(boneWeightType==0)//BDEF
	{
		my_BoneMatrix0=unpackMatrix(iBone0);
		vertex.ecPos = normalize(vec3(gl_ModelViewMatrix * my_Vertex));
		vertex.normal = normalize(gl_NormalMatrix * mat3(my_BoneMatrix0) * vertex.normal);
		return gl_ModelViewMatrix*my_BoneMatrix0*my_Vertex;
	}
	else if(boneWeightType==1)//BDEF2
	{
		//BDEF2
		my_BoneMatrix0=unpackMatrix(iBone0);
		my_BoneMatrix1=unpackMatrix(iBone1);
		tempMatrix = 
		(
			my_BoneMatrix0*fBone0Weight+
			my_BoneMatrix1*(fBone1Weight)
			);
		vertex.ecPos = normalize(vec3(gl_ModelViewMatrix * my_Vertex));
		vertex.normal = normalize(gl_NormalMatrix * mat3(tempMatrix) * vertex.normal);
		//return my_Vertex * tempMatrix * gl_ModelViewMatrix;
		return gl_ModelViewMatrix * (tempMatrix * my_Vertex);
	}
	else if(boneWeightType==2)//BDEF4
	{
		//BDEF4
		my_BoneMatrix0=unpackMatrix(iBone0);
		my_BoneMatrix1=unpackMatrix(iBone1);
		my_BoneMatrix2=unpackMatrix(iBone2);
		my_BoneMatrix3=unpackMatrix(iBone3);
		tempMatrix = 
		(
			my_BoneMatrix0*fBone0Weight+
			my_BoneMatrix1*fBone1Weight+
			my_BoneMatrix2*fBone2Weight+
			my_BoneMatrix3*fBone3Weight
			);
		vertex.ecPos = normalize(vec3(gl_ModelViewMatrix * my_Vertex));
		vertex.normal = normalize(gl_NormalMatrix * mat3(tempMatrix) * vertex.normal);
		//return my_Vertex * tempMatrix * gl_ModelViewMatrix;
		return gl_ModelViewMatrix * (tempMatrix * my_Vertex);
	}
	else if(boneWeightType==3)//SDEF
	{
		//SDEF
		my_BoneMatrix0=unpackMatrix(iBone0);
		my_BoneMatrix1=unpackMatrix(iBone1);
		vec3 sdefRW0 = sdefR0;
		vec3 sdefRW1 = sdefR1;
		vec3 vector = vec3(my_BoneMatrix0*vec4(sdefRW0,0));
		vec3 vector2 = vec3(my_BoneMatrix1*vec4(sdefRW1,0));
		vec3 vector3 = vec3(my_BoneMatrix0*vec4(sdefC,1.0));
		vec3 vector4 = vec3(my_BoneMatrix1*vec4(sdefC,1.0));
		vec3 vector5 = (fBone0Weight * ((vector3 - sdefC) + (0.5 * vector))
		             + (fBone1Weight) * ((vector4 - sdefC) + (0.5 * vector2))) 
					 + sdefC;
		mat4 matrix = quaternionToMatrix(slerp(matrixToQuaternion(my_BoneMatrix0),matrixToQuaternion(my_BoneMatrix1),fBone1Weight));
		tempMatrix = (my_BoneMatrix0*fBone0Weight+my_BoneMatrix1*(fBone1Weight));
		vertex.ecPos = normalize(vec3(gl_ModelViewMatrix * my_Vertex));
		//vertex.normal = normalize(gl_NormalMatrix * mat3(tempMatrix) * vertex.normal);
		vertex.normal = normalize(gl_NormalMatrix * mat3(matrix) * vertex.normal);
		return gl_ModelViewMatrix *vec4((vec3(matrix*vec4(vPosition-sdefC,0.0))) + vector5,1.0);
	}
	else
	{
		//BDEF2
		if(fBone0Weight>0.99)
		{
			my_BoneMatrix0=unpackMatrix(iBone0);
			vertex.ecPos = normalize(vec3(gl_ModelViewMatrix * my_Vertex));
			vertex.normal = normalize(gl_NormalMatrix * mat3(my_BoneMatrix0) * vertex.normal);
			return gl_ModelViewMatrix*my_BoneMatrix0*my_Vertex;
		}
		else if(fBone0Weight<0.01)
		{
			my_BoneMatrix1=unpackMatrix(iBone1);
			vertex.ecPos = normalize(vec3(gl_ModelViewMatrix * my_Vertex));
			vertex.normal = normalize(gl_NormalMatrix * mat3(my_BoneMatrix1) * vertex.normal);
			return gl_ModelViewMatrix*my_BoneMatrix1*my_Vertex;
		}
		else
		{
			my_BoneMatrix0=unpackMatrix(iBone0);
			my_BoneMatrix1=unpackMatrix(iBone1);
			tempMatrix = 
			(
				my_BoneMatrix0*fBone0Weight+
				my_BoneMatrix1*(fBone1Weight)
				);
			vertex.ecPos = normalize(vec3(gl_ModelViewMatrix * my_Vertex));
			vertex.normal = normalize(gl_NormalMatrix * mat3(tempMatrix) * vertex.normal);
			return gl_ModelViewMatrix * tempMatrix * my_Vertex;
		}
	}
}
